/*
 * Copyright 2006 Eric Bodden
 */
package ca.mcgill.sable.ft.reccache.cache;

import java.lang.reflect.Field;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Stack;
import java.util.Map.Entry;

import org.aspectj.lang.annotation.SuppressAjWarnings;
import org.aspectj.lang.reflect.FieldSignature;

import ca.mcgill.sable.ft.reccache.Scope;
import ca.mcgill.sable.ft.reccache.CheckpointingException;

/**
 * An aspect that implements checkpointing using a recovery cache.
 *
 * @author Eric Bodden
 */
public aspect Checkpoints {
	
	/** the actual recovery cache in form of a thread-local stack of check points */
	protected static ThreadLocal<Stack<Checkpoint>> thisThreadsCheckpoints =
		new ThreadLocal<Stack<Checkpoint>>() {		
		/** initializes to the empty stack */
		protected Stack<Checkpoint> initialValue() {
			return new Stack<Checkpoint>();
		};
	};

	// ========= Pointcuts to capture field accesses ==========
	
	/** matches when a checkpoint is active (i.e. on the stack) */
	pointcut active(): if(!thisThreadsCheckpoints.get().isEmpty());
	
	/** matches set-joinpoints of static fields,
	 * but only if a check point is active */
	pointcut staticFieldSet(): set(* *) && !this(Object) && active();

	/** matches set-joinpoints of non-static fields exposing owner object of the field,
	 * but only if a check point is active */
	pointcut fieldSet(Object owner): set(* *) && this(owner) && active();
		
	// ========= Advice to capture field accesses ==========

	/**
	 * On every non-static field set, stores the old value of the given field.
	 * @param owner the owner object of the field
	 */
	@SuppressAjWarnings("adviceDidNotMatch")
	before(Object owner): fieldSet(owner) && Scope.scope() {
		FieldSignature id = (FieldSignature)thisJoinPointStaticPart.getSignature();
		store(owner,id);
	}

	/**
	 * On every static field set, stores the old value of the given field.
	 */
	@SuppressAjWarnings("adviceDidNotMatch")
	before(): staticFieldSet() && Scope.scope() {
		FieldSignature id = (FieldSignature)thisJoinPointStaticPart.getSignature();
		Class owner = thisJoinPointStaticPart.getSignature().getDeclaringType();
		store(owner,id);
	}

	// ========= public API ==========
	
	/**
	 * Establishes a checkpoint. All subsequent changes to fields will be recorded.
	 */
	public static void establish() {
		thisThreadsCheckpoints.get().push(new Checkpoint());		
	}

	/**
	 * Restores the current state from the last checkpoint.
	 * @throws CheckpointingException if no checkpoint currently exists 
	 */
	public static void restore() {
		if(thisThreadsCheckpoints.get().isEmpty()) {
			throw new CheckpointingException("Level violation. restore() called but no checkpoint present.");
		}
		
		//pop the current checkpoint
		Checkpoint currentCheckpoint = thisThreadsCheckpoints.get().pop();
		//restore
		for (Iterator<Entry<FieldId,Object>> entryIter = currentCheckpoint.state.entrySet().iterator();
		     entryIter.hasNext();) {
			Entry<FieldId, Object> entry = entryIter.next();
			FieldId pair = entry.getKey();
			setValue(pair.o,pair.s.getField(),entry.getValue());
		}
	}
	
	/**
	 * Discards the current checkpoint. 
	 * @throws CheckpointingException if no checkpoint currently exists 
	 */
	public static void discard() {
		Stack<Checkpoint> checkpoints = thisThreadsCheckpoints.get();
		if(checkpoints.isEmpty()) {
			throw new CheckpointingException("Level violation. restore() called but no checkpoint present.");
		}

		//pop the current checkpoint
		Checkpoint currentCheckpoint = checkpoints.pop();
		
		//merge
		if(!checkpoints.isEmpty()) {
			Checkpoint enclosingCheckpoint = checkpoints.peek();
			HashMap<FieldId,Object> enclosingRestorePoint = enclosingCheckpoint.state;
			
			for (Iterator keyITer = currentCheckpoint.state.keySet().iterator(); keyITer.hasNext();) {
				FieldId pair = (FieldId) keyITer.next();
				if(!enclosingRestorePoint.containsKey(pair)) {
					enclosingRestorePoint.put(pair,currentCheckpoint.state.get(pair));
				}
			}
		}
	}
	
	// ========= helper methods for field access (reflection) ==========
	
	/**
	 * Stores the current value of the field with the given signature and owned by the
	 * given object.  
	 * @param owner the object owning that field or the class owning the field for static fields
	 * @param id the signature of the field
	 * @throws CheckpointingException if the field cannot be accessed
	 */
	protected void store(Object owner, FieldSignature id) {
		Object currValue = currentValue(owner,id);
		FieldId pair = new FieldId(owner,id);
		thisThreadsCheckpoints.get().peek().state.put(pair,currValue);
	}
	
	/**
	 * Returns the current value of the field identified by its owner object and field signature. 
	 * @param owner the object owning that field or the class owning the field for static fields
	 * @param id the signature of the field
	 * @throws CheckpointingException if the field cannot be accessed
	 */
	protected Object currentValue(Object owner, FieldSignature signature) {
		Field field;
		try {
			field = signature.getDeclaringType().getDeclaredField(signature.getName());
			field.setAccessible(true);
			return field.get(owner);
		} catch (SecurityException e) {
			throw new CheckpointingException("An exception occured when reading field "+signature+".",e);
		} catch (NoSuchFieldException e) {
			throw new CheckpointingException("An exception occured when reading field "+signature+".",e);
		} catch (IllegalArgumentException e) {
			throw new CheckpointingException("An exception occured when reading field "+signature+".",e);
		} catch (IllegalAccessException e) {
			throw new CheckpointingException("An exception occured when reading field "+signature+".",e);
		}
	}
	
	/**
	 * Sets the value of the field identified by its owner object and field signature to the given value. 
	 * @param owner the object owning that field or the class owning the field for static fields
	 * @param id the signature of the field
	 * @param value the new value
	 * @throws CheckpointingException if the field cannot be accessed or the value is of an incorrect type
	 */
	protected static void setValue(Object owner, Field field, Object value) {
		field.setAccessible(true);
		try {
			field.set(owner,value);
		} catch (IllegalArgumentException e) {
			throw new CheckpointingException("An exception occured when setting field "+field+".",e);
		} catch (IllegalAccessException e) {
			throw new CheckpointingException("An exception occured when setting field "+field+".",e);
		}
	}

	/**
	 * A single checkpoint, implemented as a mapping from field IDs to objects (values). 
	 *
	 * @author Eric Bodden
	 */
	protected static class Checkpoint {

		/** a map from a field ID to its object value */
		protected HashMap<FieldId,Object> state = new HashMap<FieldId,Object>(); 

	}
		
}
